import base64
import json
import os
import shutil
import hashlib
import httplib2
from enum import Enum
from watchdog.observers.polling import PollingObserver as Observer
from watchdog.events import FileSystemEventHandler
from common import logger, config

log = logger.getLogger('imageService')
sourceDirectory = config.getString('IMAGES_SERVICE', 'sourceDirectory', None)
destinationDirectory = "/var/nginx/images"
maxImageSize = config.getInt('IMAGES_SERVICE', 'maxImageSize', 1048576)
availableFileExtensions = {'.png', '.jpg', '.gif'}

exchangerLogin = config.getString('REST_UPLOAD', 'login')
exchangerPassword = config.getString('REST_UPLOAD', 'passwd')
exchangerHost = 'localhost'
exchangerPort = config.getInt('REST_UPLOAD', 'port', 28080)
chunkSize = config.getInt('FILEQUEUE', 'queueChunkSize', 25000)


def start():
    log.info('Запуск сервиса для синхронизации изображений')
    if sourceDirectory is None or not os.path.isdir(sourceDirectory):
        log.error('Путь к директории с изображениями не указан, либо не существует')
        return

    sync_initial()

    eventHandler = SyncEventHandler()
    observer = Observer()
    observer.schedule(eventHandler, sourceDirectory, recursive=True)
    observer.start()
    try:
        while observer.is_alive():
            observer.join(1)
    finally:
        observer.stop()
        observer.join()
        log.info('Остановка сервиса для синхронизации изображений')


def sync_initial():
    log.debug('Старт начальной синхронизации')
    if os.path.exists(destinationDirectory):
        shutil.rmtree(destinationDirectory)
    commands = []
    for command in generate_dict_commands():
        commands.append(command)
        if len(commands) >= chunkSize:
            send_commands(commands)
            commands.clear()
    if commands:
        send_commands(commands)
    log.debug('Завершение начальной синхронизации')


def generate_dict_commands():
    yield create_clear_image_command()
    for root, dirs, files in os.walk(sourceDirectory):
        for file in files:
            srcFilePath = os.path.join(root, file)
            if not is_valid_file(srcFilePath):
                continue
            relativePath = os.path.relpath(srcFilePath, sourceDirectory)
            destFilePath = modify_file_path(os.path.join(destinationDirectory, relativePath))
            if is_valid_path(relativePath):
                copy_file(srcFilePath, destFilePath)
                yield create_add_image_command(destFilePath)


def send_command(command):
    send_commands([command])


def send_commands(commands):
    try:
        http = httplib2.Http(".cache")
        base64string = base64.b64encode(('%s:%s' % (exchangerLogin, exchangerPassword)).encode()).decode()
        authHeader = "Basic %s" % base64string
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': authHeader
        }
        method = 'POST'
        url = 'http://%s:%s/aifexchange/loaddicts' % (exchangerHost, exchangerPort)

        body = json.dumps(commands)
        log.info('Выполнение загрузки справочника по url:%s' % url)
        response, content = http.request(url, method, headers=headers, body=body)
        if response.status == 200:
            content_json = json.loads(content)
            if not content_json['success']:
                log.warn('Ошибка. %s - %s' % (response, content))
        else:
            log.warn('Ошибка загрузки справочников')
    except Exception as e:
        log.error('Ошибка при отправке команд: %s', e)


def create_add_image_command(filePath):
    usageType, typeId = get_usage_type_with_type_id_from_path(filePath)
    fileName = os.path.basename(filePath)
    id = os.path.splitext(fileName)[0]
    hash = calculate_hash(filePath)
    obj = {
        'command': 'addImage',
        'image': {
            'id': id,
            'idType': typeId,
            'usageType': usageType,
            'hash': hash
        }
    }
    return obj


def create_delete_image_command(filePath):
    usageType, typeId = get_usage_type_with_type_id_from_path(filePath)
    fileName = os.path.basename(filePath)
    id = os.path.splitext(fileName)[0]
    obj = {
        'command': 'deleteImage',
        'image': {
            'id': id,
            'idType': typeId,
            'usageType': usageType
        }
    }
    return obj


def create_clear_image_command():
    return {'command': 'clearImage'}


def calculate_hash(filePath):
    hash = hashlib.blake2b()
    with open(filePath, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash.update(chunk)
    return hash.hexdigest()


def is_valid_path(path):
    parts = path.split(os.sep)
    if len(parts) < 3:
        return False
    usageType = parts[-3]
    typeId = parts[-2]
    isValidPath = usageType.upper() in UsageType._member_names_ and typeId.upper() in TypeId._member_names_
    if not isValidPath:
        log.debug('Некорректный путь: \'%s\' ' % path)
    return isValidPath


def get_usage_type_with_type_id_from_path(path):
    parts = path.split(os.sep)
    usageType = parts[-3]
    typeId = parts[-2]
    return UsageType[usageType.upper()].value, TypeId[typeId.upper()].value


def modify_file_path(path):
    parts = path.split(os.sep)
    directories = parts[:-1]
    filename = parts[-1]
    lower_directories = [dir.lower() for dir in directories]
    pathWithExtension = os.path.join(os.sep, *lower_directories, filename)
    return os.path.splitext(pathWithExtension)[0]


def copy_file(srcPath, destPath):
    os.makedirs(os.path.dirname(destPath), exist_ok=True)
    shutil.copy(srcPath, destPath)


def is_valid_file(filePath):
    fileExtension = os.path.splitext(os.path.basename(filePath))[1]
    if fileExtension not in availableFileExtensions:
        log.warn('Не удалось копировать файл \'%s\'. Неверное расширение файла. '
                 'Допустимые расширения: ( .png | .jpg | .gif )' % filePath)
        return False
    if os.path.getsize(filePath) >= maxImageSize:
        log.warn('Не удалось копировать файл \'%s\'. Размер файла превышает %s байт' % (filePath, maxImageSize))
        return False
    return True


class SyncEventHandler(FileSystemEventHandler):
    def on_modified(self, event):
        if not event.is_directory and is_valid_file(event.src_path):
            try:
                log.debug('Обработка события изменения файла \'%s\'' % event.src_path)
                relativePath = os.path.relpath(event.src_path, sourceDirectory)
                destPath = modify_file_path(os.path.join(destinationDirectory, relativePath))
                if is_valid_path(relativePath):
                    copy_file(event.src_path, destPath)
                    addCommand = create_add_image_command(destPath)
                    send_command(addCommand)
                log.debug('Изменение файла \'%s\' прошло успешно' % event.src_path)
            except Exception as e:
                log.error(u'Ошибка при изменении изображения в локальной директории: %s' % e)

    def on_created(self, event):
        if not event.is_directory and is_valid_file(event.src_path):
            try:
                log.debug('Обработка события добавления файла \'%s\'' % event.src_path)
                relativePath = os.path.relpath(event.src_path, sourceDirectory)
                destPath = modify_file_path(os.path.join(destinationDirectory, relativePath))
                if is_valid_path(relativePath):
                    copy_file(event.src_path, destPath)
                    addCommand = create_add_image_command(destPath)
                    send_command(addCommand)
                log.debug('Добавление файла \'%s\' прошло успешно' % event.src_path)
            except Exception as e:
                log.error(u'Ошибка при добавлении изображения в локальной директории: %s' % e)

    def on_deleted(self, event):
        if not event.is_directory:
            try:
                log.debug('Обработка события удаления файла \'%s\'' % event.src_path)
                relativePath = os.path.relpath(event.src_path, sourceDirectory)
                destPath = modify_file_path(os.path.join(destinationDirectory, relativePath))
                if os.path.exists(destPath):
                    deleteCommand = create_delete_image_command(destPath)
                    send_command(deleteCommand)
                    os.remove(destPath)
                log.debug('Удаление файла \'%s\' прошло успешно' % event.src_path)
            except Exception as e:
                log.error(u'Ошибка при удалении изображения из локальной директории: %s' % e)

    def on_moved(self, event):
        if not event.is_directory:
            try:
                log.debug('Обработка события перемещения файла \'%s\'' % event.src_path)
                srcRelativePath = os.path.relpath(event.src_path, sourceDirectory)
                srcFinalPath = modify_file_path(os.path.join(destinationDirectory, srcRelativePath))
                destRelativePath = os.path.relpath(event.dest_path, sourceDirectory)
                destFinalPath = modify_file_path(os.path.join(destinationDirectory, destRelativePath))
                if is_valid_path(destRelativePath):
                    copy_file(event.dest_path, destFinalPath)
                    addCommand = create_add_image_command(destFinalPath)
                    send_command(addCommand)
                    if is_valid_path(srcRelativePath) and os.path.exists(srcFinalPath):
                        deleteCommand = create_delete_image_command(srcFinalPath)
                        send_command(deleteCommand)
                        os.remove(srcFinalPath)
                log.debug('Перемещение файла \'%s\' прошло успешно' % event.src_path)
            except Exception as e:
                log.error(u'Ошибка при перемещении изображения: %s' % e)


class TypeId(Enum):
    BARCODE = 0
    CODE = 1
    ARTICUL = 2


class UsageType(Enum):
    BASIC = 0
    SECOND_SCREEN = 1
